package be.cylab.java.roc;

import com.opencsv.CSVReader;
import org.knowm.xchart.BitmapEncoder;
import org.knowm.xchart.XYChart;
import org.knowm.xchart.XYChartBuilder;
import org.knowm.xchart.XYSeries;
import org.knowm.xchart.style.lines.SeriesLines;
import org.knowm.xchart.style.markers.SeriesMarkers;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;

/**
 * Class to compute Precision-Recall curve.
 */
public class PRCurve {
    private List<Point> points;

    /**
     * @param score
     * @param true_alert
     */
    public PRCurve(final double[] score, final double[] true_alert) {
        points = new ArrayList<>();
        if (score.length != true_alert.length) {
            throw new IllegalArgumentException(
                    "Score array and true alert array must have the same size");
        }
        for (int i = 0; i < score.length; i++) {
            Point point = new Point(score[i], true_alert[i]);
            points.add(point);
        }
    }

    /**
     *
     * @param filename
     */
    public PRCurve(final String filename) {
        points = new ArrayList<>();
        CSVReader reader;
        try {
            reader = new CSVReader(new InputStreamReader(
                    new FileInputStream(filename), StandardCharsets.UTF_8));
            String[] next_record = null;
            while (true) {
                try {
                    next_record = reader.readNext();
                    if (!(next_record != null)) {
                        break;
                    }
                } catch (IOException e) {
                    e.printStackTrace();
                }
                Point point = new Point(Double.parseDouble(next_record[0]),
                        Double.parseDouble(next_record[1]));
                points.add(point);
            }

        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }
    }

    /**
     * Method to compute points in the PR space.
     * @return
     */
    public final List<CurveCoordinates> computePrcPoints() {
        double iteration = 0.001;
        double threshold = 1.0;
        List<CurveCoordinates> curve_points = new ArrayList<>();
        Statistics actual = new Statistics(points, threshold);
        Statistics next = new Statistics(points, threshold - iteration);
        while (Double.isNaN(actual.getPrecision())
                && Double.isNaN(next.getPrecision())) {
            threshold = threshold - iteration;
            actual = new Statistics(points, threshold);
            next = new Statistics(points, threshold - iteration);
        }
        if (Double.isNaN(actual.getPrecision())) {
            if (next.getTruePositiveCounter() == 0) {
                CurveCoordinates point = new CurveCoordinates(
                        next.getRecall(), next.getPrecision());
                curve_points.add(point);
            } else {
                actual.setPrecision(next.getPrecision());
                CurveCoordinates p1 = new CurveCoordinates(
                        actual.getRecall(), actual.getPrecision());
                CurveCoordinates p2 = new CurveCoordinates(
                        next.getRecall(), next.getPrecision());
                curve_points.add(p1);
                curve_points.add(p2);
            }
        }
        for (double i = threshold - iteration; i > 0.0; i -= iteration) {
            actual = new Statistics(points, i);
            next = new Statistics(points, i - iteration);
            CurveCoordinates inter = Utils.interpolationPoint(actual, next);
            CurveCoordinates a = new CurveCoordinates(actual.getRecall(),
                    actual.getPrecision());
            CurveCoordinates n = new CurveCoordinates(next.getRecall(),
                    next.getPrecision());
            curve_points.add(a);
            curve_points.add(inter);
            curve_points.add(n);
        }
        return curve_points;
    }

    /**
     * Method to generate th P-R curve and save in a png file.
     * @param filename
     */
    public final void generatePrCurve(final String filename) {
        List<CurveCoordinates> pr_coordinates = computePrcPoints();
        double[] precision = new double[pr_coordinates.size()];
        double[] recall = new double[pr_coordinates.size()];
        //Create arrays to store Recall-Precision coordinates
        for (int i = 0; i < pr_coordinates.size(); i++) {
            precision[i] = pr_coordinates.get(i).getYAxis();
            recall[i] = pr_coordinates.get(i).getXAxis();
        }
        //Chart creation
        XYChart chart = new XYChartBuilder().width(600).height(400)
                .title("PR-Curve").xAxisTitle("Recall").yAxisTitle("Precision")
                .build();

        //P / (P + N) computation
        int pos_ex = Utils.countPositiveExamples(points);
        int neg_ex = Utils.countNegativeExamples(points);
        double y_dotted_position = pos_ex / (double) (pos_ex + neg_ex);
        //Set up graph to visualize random classification
        double[] dotted_x = new double[] {0, 1};
        double[] dotted_y = new double[] {y_dotted_position, y_dotted_position};
        XYSeries dotted = chart.addSeries("Random Classification",
                dotted_x, dotted_y);
        dotted.setLineStyle(SeriesLines.DASH_DASH);
        dotted.setMarker(SeriesMarkers.NONE);

        //Set up graph to visualize P-R curve
        XYSeries pr = chart.addSeries("P-R curve", recall, precision);
        pr.setMarker(SeriesMarkers.NONE);
        //Save graph as PNG
        try {
            BitmapEncoder.saveBitmapWithDPI(chart,
                    filename,
                    BitmapEncoder.BitmapFormat.PNG,
                    300);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * @return
     */
    public final double computePRAUC() {
        List<CurveCoordinates> pr_coordinates = computePrcPoints();
        CurveCoordinates previous;
        CurveCoordinates actual;
        double area = 0;
        for (int i = 1; i < pr_coordinates.size() - 1; i++) {
            previous = pr_coordinates.get(i - 1);
            actual = pr_coordinates.get(i);
            area = area + trapezoidArea(previous, actual);
        }
        return area;
    }


    /**
     * @param previous
     * @param actual
     * @return
     */
    private double trapezoidArea(final CurveCoordinates previous,
                                 final CurveCoordinates actual) {
        double b1 = previous.getYAxis();
        double b2 = actual.getYAxis();
        double height = Math.abs(actual.getXAxis() - previous.getXAxis());
        return (((b1 + b2) * height) / 2);
    }

}
